from datetime import datetime, timedelta
from functools import reduce
from typing import List, Tuple

import numpy as np
import pandas as pd

from stratestic.backtesting.helpers import Trade


def geometric_mean(returns: pd.Series) -> float:
    returns = returns.fillna(0) + 1

    if np.any(returns <= 0):
        return np.nan

    return np.exp(np.log(returns).sum() / (len(returns) or np.nan)) - 1


def get_total_duration(index: np.ndarray) -> timedelta:
    return index[-1] - index[0]


def get_start_date(index: np.ndarray) -> datetime:
    return index[0]


def get_end_date(index: np.ndarray) -> datetime:
    return index[-1]


def exposure_time(positions: np.ndarray) -> float:
    """Calculate the percentage of time the strategy was exposed to the market."""
    return np.count_nonzero(positions) / len(positions) * 100


def equity_final(equity_curve: pd.Series) -> float:
    """Calculate the final equity value."""
    return equity_curve.iloc[-1]


def equity_peak(equity_curve: pd.Series) -> float:
    """Calculate the peak equity value."""
    return np.max(equity_curve)


def return_pct(equity_curve: pd.Series) -> float:
    """Calculate the total return percentage."""
    return (equity_curve.iloc[-1] / equity_curve.iloc[0] - 1) * 100


def return_buy_and_hold_pct(returns: pd.Series) -> float:
    """Retrieve the buy and hold return pct."""
    return (returns[-1] - 1) * 100


def return_pct_annualized(cum_returns: pd.Series, leverage=1) -> float:
    """Calculate the annualized return percentage."""

    years_df = cum_returns.resample('1Y').count()

    return ((cum_returns[-1] / cum_returns[0] * leverage)**(1 / len(years_df))-1) * 100


def volatility_pct_annualized(returns: pd.Series, trading_days: int = 365) -> float:
    """Calculate the annualized volatility percentage."""

    days_df = returns.resample('1D').sum()

    return np.std(days_df) * np.sqrt(trading_days) * 100


def sharpe_ratio(returns: pd.Series, risk_free_rate: float = 0, trading_days: int = 365) -> float:
    """Calculate the Sharpe ratio."""
    excess_returns = returns - risk_free_rate
    excess_returns_dev = np.std(excess_returns)

    return np.mean(excess_returns) / excess_returns_dev * np.sqrt(trading_days) if excess_returns_dev != 0 else np.nan


def sortino_ratio(returns: pd.Series, target_return: float = 0, risk_free_rate: float = 0) -> float:
    """Calculate the Sortino ratio."""
    downside_returns = returns.copy()
    downside_returns[downside_returns > target_return] = 0
    downside_deviation = volatility_pct_annualized(downside_returns)
    cumulative_returns = returns.cumsum().apply(np.exp)
    sortino_ratio = (return_pct_annualized(cumulative_returns) - risk_free_rate) / downside_deviation
    return sortino_ratio


def calmar_ratio(cum_returns: pd.Series, risk_free_rate: float = 0) -> float:
    """Calculate the Calmar ratio."""
    max_drawdown = max_drawdown_pct(cum_returns) / 100
    annual_return = return_pct_annualized(cum_returns) / 100
    return (annual_return - risk_free_rate) / -max_drawdown


def get_drawdowns(cum_returns: pd.Series) -> pd.Series:
    """Retrieves the drawdown periods."""
    peaks = np.maximum.accumulate(cum_returns)
    drawdowns = (cum_returns - peaks) / peaks

    return drawdowns


def max_drawdown_pct(cum_returns: pd.Series) -> float:
    """Calculate the maximum drawdown percentage."""
    drawdowns = get_drawdowns(cum_returns)

    return np.min(drawdowns) * 100


def avg_drawdown_pct(cum_returns: pd.Series) -> float:
    """Calculate the average drawdown percentage."""
    drawdowns = get_drawdowns(cum_returns)

    drawdown_maximums = []
    drawdown_period = []

    is_zero = True
    for dd in drawdowns:
        if dd == 0:
            if not is_zero:
                if len(drawdown_period) != 0:
                    drawdown_maximums.append(np.min(drawdown_period))
            is_zero = True
        else:
            if is_zero:
                drawdown_period = []
                is_zero = False
            else:
                drawdown_period.append(dd)

    if len(drawdown_period) != 0:
        drawdown_maximums.append(np.min(drawdown_period))

    return np.mean(drawdown_maximums) * 100


def max_drawdown_duration(cum_returns: pd.Series) -> int:
    """Calculate the duration of the maximum drawdown in terms of the period between peak to peak.

    Parameters
    ----------
    cum_returns : pd.Series
        An array of cumulative returns.

    Returns
    -------
    int
        The duration of the maximum drawdown between peak to peak in the used time unit.
    """
    peak_index = 0
    max_drawdown_dur = 0
    dd_start = cum_returns.index[0]
    dd_end = cum_returns.index[-1]

    for i in range(1, len(cum_returns)):
        if cum_returns[i] > cum_returns[peak_index]:
            peak_index = i
        else:
            drawdown_duration = i - peak_index
            if drawdown_duration > max_drawdown_dur:
                max_drawdown_dur = drawdown_duration
                dd_start = cum_returns.index[peak_index]
                dd_end = cum_returns.index[i]

    return dd_end - dd_start


def get_dd_durations_limits(cum_returns: pd.Series) -> Tuple[List, List]:
    peak_index = 0

    durations = []
    limits = []

    drawdown = False
    dd_duration = 0
    dd_start_end = (None, None)
    dd_start = cum_returns.index[0]
    for i in range(1, len(cum_returns)):
        if cum_returns[i] > cum_returns[peak_index]:
            dd_start = cum_returns.index[i]
            peak_index = i

            if drawdown:
                durations.append(dd_duration)
                limits.append(dd_start_end)
                drawdown = False
        else:
            drawdown = True
            dd_end = cum_returns.index[i]
            dd_start_end = (dd_start, dd_end)
            dd_duration = (dd_end - dd_start).total_seconds()

    if drawdown:
        durations.append(dd_duration)
        limits.append(dd_start_end)

    return durations, limits


def avg_drawdown_duration(cum_returns: pd.Series) -> float:
    """Calculate the average duration of drawdowns."""
    durations, _ = get_dd_durations_limits(cum_returns)

    return np.mean(durations) if len(durations) > 0 else 0


def win_rate_pct(trades: List[Trade]) -> float:
    """Calculate the percentage of winning trades."""
    winning_trades = reduce(
        lambda accum, trade: accum + 1 if (trade.exit_price - trade.entry_price) * trade.side > 0 else accum,
        trades,
        0
    )

    nr_trades = len(trades)

    return winning_trades / nr_trades * 100 if nr_trades > 0 else 0


def calculate_pnl(trade):
    return np.exp(np.log(trade.exit_price / trade.entry_price) * trade.side) - 1


def best_trade_pct(trades: List[Trade], leverage=1) -> float:
    """Calculate the percentage of the best trade."""

    def reducer(accum, trade):
        pnl = calculate_pnl(trade)
        return pnl if pnl > accum else accum

    best_trade = reduce(
        reducer,
        trades,
        0
    )
    return best_trade * leverage * 100


def worst_trade_pct(trades: List[Trade], leverage=1) -> float:
    """Calculate the percentage of the worst trade."""

    def reducer(accum, trade):
        pnl = calculate_pnl(trade)
        return pnl if pnl < accum else accum

    worst_trade = reduce(
        reducer,
        trades,
        0
    )

    return worst_trade * leverage * 100


def avg_trade_pct(trades: List[Trade], leverage=1) -> float:
    """Calculate the average trade percentage."""
    trades_pct = map(
        # lambda trade: (trade.exit_price - trade.entry_price) * leverage / trade.entry_price * trade.side,
        lambda trade: calculate_pnl(trade) * leverage,
        trades
    )

    intermediate = pd.Series(list(trades_pct))

    return geometric_mean(intermediate) * 100


def max_trade_duration(trades: List[Trade]) -> int:
    """Calculate the duration of the longest trade."""
    durations = list(map(
        lambda trade: trade.exit_date - trade.entry_date,
        trades,
    ))

    return np.max(durations) if len(durations) > 0 else 0


def avg_trade_duration(trades: List[Trade]) -> int:
    """Calculate the duration of the longest trade."""
    durations = map(
        lambda trade: (trade.exit_date - trade.entry_date).total_seconds(),
        trades,
    )

    durations = list(durations)

    return np.mean(durations) if len(durations) > 0 else 0


def winning_trades(trades: List[Trade]) -> List[Trade]:
    return reduce(
        lambda accum, trade: [*accum, trade]
        if (trade.exit_price - trade.entry_price) * trade.side > 0
        else accum,
        trades,
        []
    )


def losing_trades(trades: List[Trade]) -> List[Trade]:
    return reduce(
        lambda accum, trade: [*accum, trade]
        if (trade.exit_price - trade.entry_price) * trade.side < 0
        else accum,
        trades,
        []
    )


def trades_net_profit_sum(trades: List[Trade]) -> float:
    return reduce(
        lambda accum, trade: accum + trade.units * (trade.exit_price - trade.entry_price) * trade.side,
        trades,
        0
    )


def trades_net_profit(trades: List[Trade]) -> List[float]:
    return list(map(
        lambda trade: trade.units * (trade.exit_price - trade.entry_price) * trade.side,
        trades,
    ))


def profit_factor(trades: List[Trade]) -> float:
    """Calculate the profit factor."""
    win_trades = winning_trades(trades)
    lose_trades = losing_trades(trades)

    if len(lose_trades) == 0:
        return np.inf
    else:
        return trades_net_profit_sum(win_trades) / np.abs(trades_net_profit_sum(lose_trades))


def expectancy_pct(trades: List[Trade]) -> float:
    """Calculate the expectancy percentage."""
    win_trades = winning_trades(trades)
    lose_trades = losing_trades(trades)
    if len(lose_trades) == 0:
        return avg_trade_pct(win_trades)
    elif len(win_trades) == 0:
        return avg_trade_pct(lose_trades)
    else:
        win_rate = win_rate_pct(trades) / 100
        avg_win = avg_trade_pct(win_trades) / 100
        avg_loss = avg_trade_pct(lose_trades) / 100
        return (win_rate * avg_win - (1 - win_rate) * avg_loss) * 100


def system_quality_number(trades):
    """Calculate the System Quality Number."""

    net_profit = trades_net_profit(trades)

    N = len(trades)
    avg = np.mean(net_profit)
    std = np.std(net_profit)

    return np.sqrt(N) * (avg / std)



import logging

from stratestic.backtesting.helpers.evaluation.metrics import *
from stratestic.backtesting.helpers.evaluation._constants import results_sections, results_mapping, results_aesthetics

import pandas as pd


def get_results(processed_data, trades, leverage, amount, trading_costs=None, trading_days=365):

    results = {}

    results["total_duration"] = get_total_duration(processed_data.index)
    results["start_date"] = get_start_date(processed_data.index)
    results["end_date"] = get_end_date(processed_data.index)

    results["leverage"] = leverage

    if trading_costs is not None:
        results["trading_costs"] = trading_costs * 100

    results["initial_equity"] = amount
    results["exposed_capital"] = results["initial_equity"] / leverage

    if "side" in processed_data:
        results["exposure_time"] = exposure_time(processed_data["side"])

    if "accumulated_returns" in processed_data:
        results["buy_and_hold_return"] = return_buy_and_hold_pct(processed_data["accumulated_returns"]) * leverage

    if "accumulated_strategy_returns_tc" in processed_data:
        results["equity_final"] = equity_final(processed_data["accumulated_strategy_returns_tc"] * amount)
        results["equity_peak"] = equity_peak(processed_data["accumulated_strategy_returns_tc"] * amount)
        results["return_pct"] = return_pct(processed_data["accumulated_strategy_returns_tc"]) * leverage
        results["return_pct_annualized"] = return_pct_annualized(processed_data["accumulated_strategy_returns_tc"], leverage)
        results["calmar_ratio"] = calmar_ratio(processed_data["accumulated_strategy_returns_tc"])
        results["max_drawdown"] = max_drawdown_pct(processed_data["accumulated_strategy_returns_tc"])
        results["avg_drawdown"] = avg_drawdown_pct(processed_data["accumulated_strategy_returns_tc"])
        results["max_drawdown_duration"] = max_drawdown_duration(processed_data["accumulated_strategy_returns_tc"])
        results["avg_drawdown_duration"] = avg_drawdown_duration(processed_data["accumulated_strategy_returns_tc"])

    if "strategy_returns_tc" in processed_data:
        results["volatility_pct_annualized"] = volatility_pct_annualized(processed_data["strategy_returns_tc"], trading_days)
        results["sharpe_ratio"] = sharpe_ratio(processed_data["strategy_returns_tc"], trading_days=trading_days)
        results["sortino_ratio"] = sortino_ratio(processed_data["strategy_returns_tc"])

    results["nr_trades"] = int(len(trades))
    results["win_rate"] = win_rate_pct(trades)
    results["best_trade"] = best_trade_pct(trades, leverage)
    results["worst_trade"] = worst_trade_pct(trades, leverage)
    results["avg_trade"] = avg_trade_pct(trades, leverage)
    results["max_trade_duration"] = max_trade_duration(trades)
    results["avg_trade_duration"] = avg_trade_duration(trades)
    results["profit_factor"] = profit_factor(trades)
    results["expectancy"] = expectancy_pct(trades)
    results["sqn"] = system_quality_number(trades)

    return pd.Series(results)


def log_results(results, backtesting=True):

    length = 50

    logging.info("")

    title = 'BACKTESTING' if backtesting else 'TRADING BOT'

    logging.info('*' * length)
    logging.info(f'{title} RESULTS'.center(length))
    logging.info('*' * length)
    logging.info('')

    for section, columns in results_sections.items():

        logging.info(section.center(length))
        logging.info('-' * length)

        for col in columns:

            value = results[col]

            if callable(results_mapping[col]):
                printed_title = results_mapping[col]('USDT')
            else:
                printed_title = results_mapping[col]

            if col in results_aesthetics:
                value = results_aesthetics[col](value)
            else:
                try:
                    value = str(round(value, 2))
                except TypeError:
                    value = str(value)

            logging.info(f'{printed_title:<25}{value.rjust(25)}')
        logging.info('-' * length)
        logging.info('')
    logging.info('*' * length)


from collections import OrderedDict

import numpy as np
import pandas as pd
from ta.trend import MACD

from stratestic.strategies._mixin import StrategyMixin


class MovingAverageConvergenceDivergence(MACD, StrategyMixin):
    """
    Strategy that uses 3 different parameters to calculate the side.
    It calculates the difference between the slow and the fast exponential moving averages,
    and compares that with the signal window to get the signal. When the signal moving average goes
    above the difference of the fast and slow, it goes long, and goes short when the reverse holds true.

    Parameters
    ----------
    window_slow : int, optional
        Slow moving average window, by default 26
    window_fast : int, optional
        Fast moving average window, by default 12
    window_sign : int, optional
        Signal moving average window, by default 9
    data : pd.DataFrame, optional
        Data to use, by default None
    **kwargs
        Keyword arguments to pass to parent class

    Attributes
    ----------
    params : collections.OrderedDict
        Parameters of the strategy
    _close : pd.Series
        Close prices of the data
    """

    def __init__(
        self,
        window_slow: int = 26,
        window_fast: int = 12,
        window_sign: int = 9,
        data: pd.DataFrame = None,
        **kwargs
    ):
        """
        Constructs all the necessary attributes for the MACD strategy object.

        Parameters
        ----------
        window_slow : int, optional
            Slow moving average window, by default 26
        window_fast : int, optional
            Fast moving average window, by default 12
        window_signal : int, optional
            Signal moving average window, by default 9
        data : pd.DataFrame, optional
            Data to use, by default None
        **kwargs
            Keyword arguments to pass to parent class
        """
        MACD.__init__(self, pd.Series(dtype='float64'), window_slow, window_fast, window_sign)
        StrategyMixin.__init__(self, data, **kwargs)

        self._close = pd.Series(dtype='float64')

        self.params = OrderedDict(
            window_slow=lambda x: int(x),
            window_fast=lambda x: int(x),
            window_sign=lambda x: int(x)
        )

    def __repr__(self) -> str:
        """
        Returns the string representation of the object.

        Returns
        -------
        str
            String representation of the object.
        """
        return "{}(symbol = {}, fast = {}, slow = {}, signal = {})".format(
            self.__class__.__name__, self.symbol, self._window_fast, self._window_slow, self._window_sign
        )

    def update_data(self, data) -> None:
        """
        Updates the input data with additional columns required for the strategy.

        Parameters
        ----------
        data : pd.DataFrame
            OHLCV data to be updated.

        Returns
        -------
        pd.DataFrame
            Updated OHLCV data containing additional columns.
        """
        data = super().update_data(data)

        self._close = data[self.close_col]
        self._run()

        data["macd_diff"] = self.macd_diff()

        return data

    def calculate_positions(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Calculates positions based on MACD difference.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe containing MACD difference column

        Returns
        -------
        pd.DataFrame
            Dataframe containing side column.
        """
        data["side"] = np.where(data["macd_diff"] > 0, 1, -1)

        return data

    def get_signal(self, row=None):
        """
        Returns signal based on current data.

        Parameters
        ----------
        row : pd.Series, optional
            Row of OHLCV data to generate signal for, by default None.

        Returns
        -------
        int
            Signal (-1 for short, 1 for long, 0 for neutral).
        """
        if row is None:
            row = self.data.iloc[-1]

        if row["macd_diff"] > 0:
            return 1
        elif row["macd_diff"] < 0:
            return -1



import numpy as np

from stratestic.strategies._mixin import StrategyMixin
from collections import OrderedDict


class Momentum(StrategyMixin):
    """
    this Momentum strategy calculates the rolling average return over a specified window of time,
    and generates trading signals based on the sign of this rolling average. When the rolling average
    return is positive, the strategy generates a long signal (i.e., buy), and when it is negative,
    it generates a short signal (i.e., sell). The strategy takes a neutral side when the rolling average is zero.

    Parameters
    ----------
    window : int
        Rolling window for computing the momentum.
    data : pd.DataFrame
        Input data, with at least a 'returns_col' column.
    **kwargs
        Additional keyword arguments.

    Attributes
    ----------
    _window : int
        Rolling window for computing the momentum.
    params : OrderedDict
        Dictionary of hyperparameters for the strategy.

    Methods
    -------
    update_data()
        Retrieves and prepares the data.
    calculate_positions(data)
        Calculates the positions of the strategy.
    get_signal(row=None)
        Returns the trading signal for a given row of data.

    """

    def __init__(self, window: int, data=None, **kwargs):

        self._window = window

        StrategyMixin.__init__(self, data, **kwargs)

        self.params = OrderedDict(window=lambda x: int(x))

    def __repr__(self):
        """
        Return a string representation of the class.
        """
        return "{}(symbol = {}, window = {})".format(self.__class__.__name__, self.symbol, self._window)

    def update_data(self, data):
        """
        Updates the input data with additional columns required for the strategy.

        Parameters
        ----------
        data : pd.DataFrame
            OHLCV data to be updated.

        Returns
        -------
        pd.DataFrame
            Updated OHLCV data containing additional columns.
        """
        data = super().update_data(data)

        data["rolling_returns"] = data[self.returns_col].rolling(self._window, min_periods=1).mean()

        return data

    def calculate_positions(self, data):
        """
        Calculates the positions of the strategy.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe with the input data.

        Returns
        -------
        pd.DataFrame
            Dataframe with the positions of the strategy.
        """
        data["side"] = np.sign(data["rolling_returns"])

        return data

    def get_signal(self, row=None):
        """
        Returns the trading signal for a given row of data.

        Parameters
        ----------
        row : pd.Series, optional
            Row of data to calculate the signal for. If None, the last row of the data is used.

        Returns
        -------
        int
            The trading signal (-1 for sell, 1 for buy, 0 for hold).
        """
        if row is None:
            row = self.data.iloc[-1]

        if row["rolling_returns"] > 0:
            return 1
        elif row["rolling_returns"] < 0:
            return -1
        else:
            return 0



import numpy as np
import pandas as pd

exception_message = 'The provided exchange is not supported.'


def get_maintenance_margin(symbol_brackets, notional_value, exchange='binance'):
    """
    Calculate the maintenance margin rate and amount based on symbol brackets and notional value.

    Parameters
    ----------
    symbol_brackets: pd.DataFrame
        DataFrame containing symbol brackets with columns ['notionalFloor', 'maintMarginRatio', 'cum'].
    notional_value: pd.Series
        Series containing notional values for calculation.
    exchange: str, optional
        Exchange name. Default is 'binance'.

    Returns:
    --------
    Tuple of maintenance rate and maintenance amount.
    """
    def get_maintenance_margin_binance():
        """
        Internal function for calculating maintenance margin for Binance exchange.
        """

        # values = pd.concat([pd.DataFrame(notional_value).T] * len(notional_value)).T
        #
        # triangular_matrix = pd.DataFrame(np.tril(values))

        values = pd.Series(notional_value)

        comparison = pd.DataFrame()
        for i, notional_floor in enumerate(symbol_brackets['notionalFloor']):
            comparison[i] = values >= notional_floor

        # greater_than = triangular_matrix.gt(symbol_brackets["notionalFloor"], axis=1)

        indexes = comparison.idxmin(axis=1) - 1

        brackets = symbol_brackets.iloc[indexes, :]

        maintenance_rate = brackets["maintMarginRatio"].values
        maintenance_amount = brackets["cum"].values

        return maintenance_rate, maintenance_amount

    if exchange == 'binance':
        return get_maintenance_margin_binance()
    else:
        raise Exception(exception_message)


def calculate_margin_ratio(
    leverage,
    units,
    side,
    entry_price,
    mark_price,
    maintenance_rate,
    maintenance_amount,
    exchange
):
    """
    Calculate the margin ratio based on provided parameters.

    Parameters
    ----------
    leverage: float
        Leverage used for the position.
    units: float
        Number of units in the position.
    side: int
        Trading side (-1 for short, 1 for long).
    entry_price: float
        Entry price of the position.
    mark_price: float
        Mark price of the position.
    maintenance_rate: float
        Maintenance margin rate.
    maintenance_amount: float
        Maintenance margin amount.
    exchange: str
        Exchange name.

    Returns:
    --------
    Float representing the margin ratio.
    """
    def calculate_margin_ratio_binance():
        """
        Internal function for calculating margin ratio for Binance exchange.
        """
        initial_value = units * entry_price
        current_value = units * mark_price
        initial_margin = initial_value / leverage

        margin_balance = initial_margin + (current_value - initial_value) * side

        maintenance_margin = current_value * maintenance_rate - maintenance_amount

        try:
            return maintenance_margin / margin_balance
        except ZeroDivisionError:
            return np.Inf

    if exchange == 'binance':
        return calculate_margin_ratio_binance()
    else:
        raise Exception(exception_message)


def calculate_liquidation_price(
    units,
    entry_price,
    side,
    leverage,
    maintenance_rate,
    maintenance_amount,
    exchange
):
    """
    Calculate the liquidation price based on provided parameters.

    Parameters
    ----------
    units: float
        Number of units in the position.
    entry_price: float
        Entry price of the position.
    side: int
        Trading side (-1 for short, 1 for long).
    leverage: float
        Leverage used for the position.
    maintenance_rate: float
        Maintenance margin rate.
    maintenance_amount: float
        Maintenance margin amount.
    exchange: str
        Exchange name.

    Returns:
    --------
    Float representing the liquidation price.
    """
    def calculate_liquidation_ratio_binance():
        """
        Internal function for calculating liquidation price ratio for Binance exchange.
        """
        notional_value = units * entry_price
        wallet_balance = notional_value / leverage

        try:
            liquidation_price = ((wallet_balance + maintenance_amount - side * units * entry_price) /
                                 (units * maintenance_rate - side * units))
        except ZeroDivisionError:
            liquidation_price = 0

        return liquidation_price

    if exchange == 'binance':
        return calculate_liquidation_ratio_binance()
    else:
        raise Exception(exception_message)



class LeverageInvalid(Exception):
    def __init__(self, *args):
        if not args:
            self.message = f"Leverage Invalid."
        else:
            self.message = f"{args[0]} is not a valid leverage."

    def __str__(self):
        return f"{self.message}"

    def __repr__(self):
        return self.__class__.__name__



import numpy as np
import pandas as pd


class StrategyMixin:
    """
    A mixin class that provides basic functionality for backtesting trading strategies.

    Parameters
    ----------
    data : pd.DataFrame, optional
        The DataFrame containing the historical price data for the asset.
    close_col : str, optional
        The name of the column in the data that contains the price data.
    returns_col : str, optional
        The name of the column in the data that will contain the returns data.

    Attributes
    ----------
    data : pd.DataFrame
        The DataFrame containing the historical price data for the asset.
    close_col : str
        The name of the column in the data that contains the price data.
    returns_col : str
        The name of the column in the data that will contain the returns data.
    symbol : str
        The name of the asset being traded.

    Methods
    -------
    _get_test_title(self)
        Return a string with the title for the backtest.
    _get_data() -> pd.DataFrame
        Returns the current DataFrame containing the historical price data for the asset.
    set_data(data: pd.DataFrame) -> None
        Sets the DataFrame containing the historical price data for the asset.
    set_parameters(params: dict, data: pd.DataFrame) -> None
        Updates the parameters of the strategy.
    _calculate_returns() -> None
        Calculates the returns of the asset and updates the data DataFrame.
    update_data() -> None
        Updates the data DataFrame by calculating the returns of the asset.

    """

    def __init__(
        self,
        data=None,
        trade_on_close=True,
        close_col='close',
        open_col='open',
        high_col='high',
        low_col='low',
        returns_col='returns'
    ):
        """
        Initializes a new instance of the StrategyMixin class.

        Parameters
        ----------
        data : pd.DataFrame, optional
            The DataFrame containing the historical price data for the asset.
        close_col : str, optional
            The name of the column in the data that contains the close price data.
        open_col : str, optional
            The name of the column in the data that contains the open price data.
        high_col : str, optional
            The name of the column in the data that contains the high price data.
        low_col : str, optional
            The name of the column in the data that contains the low price data.
        returns_col : str, optional
            The name of the column in the data that will contain the returns' data.
        """

        self.close_col = close_col
        self.open_col = open_col
        self.high_col = high_col
        self.low_col = low_col
        self.trade_on_close = trade_on_close
        self.price_col = close_col if trade_on_close else open_col
        self.returns_col = returns_col
        self.symbol = None

        if data is not None:
            self.data = self.update_data(data.copy())

    def __repr__(self):
        """
        Returns a string representation of the strategy.

        Returns:
        --------
        str:
            A string representation of the strategy.
        """
        return "{}".format(self.__class__.__name__)

    def get_params(self, **kwargs):
        return self.params if self.params else {}

    def _get_test_title(self):
        """
        Returns the title for the backtest report.

        Returns:
        --------
        str:
            The title for the backtest report.
        """
        return f"{self.__repr__()} strategy backtest."

    def _get_data(self) -> pd.DataFrame:
        """
        Returns the current DataFrame containing the historical price data for the asset.

        Returns
        -------
        pd.DataFrame
            The current DataFrame containing the historical price data for the asset.
        """

        return self.data

    def set_data(self, data: pd.DataFrame, strategy_obj=None) -> None:
        """
        Sets the DataFrame containing the historical price data for the asset.

        Parameters
        ----------
        data : pd.DataFrame
            The DataFrame containing the historical price data for the asset.
        strategy_obj : Strategy object

        """

        if data is not None:
            if strategy_obj is not None:
                strategy_obj.data = strategy_obj.update_data(data.copy())
            else:
                self.data = self.update_data(self.data.copy())

    def set_parameters(self, params=None, data=None) -> None:
        """
        Updates the parameters of the strategy.

        Parameters
        ----------
        data
        params : dict, optional
            A dictionary containing the parameters to be updated.
        """

        if params is None:
            return

        for param, new_value in params.items():
            setattr(self, f"_{param}", self.get_params()[param](new_value))

        data = data.copy() if data is not None else self.data.copy()

        self.data = self.update_data(data)

    def _calculate_returns(self, data) -> pd.DataFrame:
        """
        Calculates the returns of the asset and updates the data DataFrame.
        """

        if self.trade_on_close:
            data[self.returns_col] = np.log(data[self.price_col] / data[self.price_col].shift(1))
        else:
            data[self.returns_col] = np.log(data[self.price_col].shift(-1) / data[self.price_col])
            data.loc[data.index[-1], self.returns_col] = \
                np.log(data.loc[data.index[-1], self.close_col] / data.loc[data.index[-1], self.open_col])

        return data

    def update_data(self, data) -> pd.DataFrame:
        """
        Updates the input data with additional columns required for the strategy.

        Parameters
        ----------
        data : pd.DataFrame
            OHLCV data to be updated.

        Returns
        -------
        pd.DataFrame
            Updated OHLCV data containing additional columns.
        """

        return self._calculate_returns(data)



from stratestic.backtesting.helpers.evaluation._results import get_results, log_results
from stratestic.backtesting.helpers.evaluation._constants import results_mapping, results_sections, results_aesthetics



from stratestic.utils.exceptions.symbol_invalid import SymbolInvalid
from stratestic.utils.exceptions.no_such_pipeline import NoSuchPipeline
from stratestic.utils.exceptions.optimization_parameters_invalid import OptimizationParametersInvalid
from stratestic.utils.exceptions.strategy_invalid import StrategyInvalid
from stratestic.utils.exceptions.strategy_required import StrategyRequired
from stratestic.utils.exceptions.leverage_invalid import LeverageInvalid
from stratestic.utils.exceptions.no_config_file import NoConfigFile



import logging
import sys


def configure_logger(level='INFO', libraries_level=None):
    logger = logging.getLogger()

    if not isinstance(libraries_level, list):
        libraries_level = []

    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    logger.setLevel(getattr(logging, level))

    for library, level in libraries_level:
        logging.getLogger(library).setLevel(getattr(logging, level))

    log_handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(log_handler)

    return logger



import os
import sys
from collections import namedtuple
from configparser import RawConfigParser

from stratestic.utils.exceptions import NoConfigFile

try:
    directory_path = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(directory_path)
except NameError:
    pass


def search_for_file(dirname, filename):

    filepath = os.path.join(dirname, filename)
    if os.path.exists(filepath):
        return filepath
    else:
        raise NoConfigFile


def get_config(app='', filename='proj.conf'):

    config_vars = {}

    for section in ['general', app]:

        filepath = search_for_file(os.path.dirname(__file__), filename)

        fp = open(filepath)
        config = RawConfigParser()
        config.read_file(fp)

        if section in config.sections():

            section_vars = dict(config.items(section))

            for section_var, value in section_vars.items():

                try:
                    value = float(value)
                except ValueError:
                    pass

                config_vars[section_var] = value

    config_vars = namedtuple('CONFIG_VARS', config_vars.keys())(*config_vars.values())

    return config_vars



import json
import logging
import math

import numpy as np
import pandas as pd
import progressbar
from scipy.optimize import brute
import plotly.io as pio

from stratestic.backtesting.combining import StrategyCombiner
from stratestic.backtesting.helpers.evaluation import get_results, log_results
from stratestic.backtesting.helpers.plotting import plot_backtest_results
from stratestic.utils.config_parser import get_config
from stratestic.utils.exceptions import StrategyRequired, OptimizationParametersInvalid, SymbolInvalid, LeverageInvalid
from stratestic.utils.logger import configure_logger

config_vars = get_config('general')

configure_logger(config_vars.logger_level)

pio.renderers.default = "browser"


class BacktestMixin:
    """A Mixin class for backtesting trading strategies.

    Attributes:
    -----------
    symbol : str
        The trading symbol used for the backtest.
    tc : float
        The transaction costs (e.g. spread, commissions) as a percentage.
    results : pandas.DataFrame
        A DataFrame containing the results of the backtest.

    Methods:
    --------
    run(params=None, print_results=True, plot_results=True, plot_positions=False):
        Runs the trading strategy and prints and/or plots the results.
    optimize(params, **kwargs):
        Optimizes the trading strategy using brute force.
    _test_strategy(params=None, print_results=True, plot_results=True, plot_positions=False):
        Tests the trading strategy on historical data.
    _assess_strategy(data, title, print_results=True, plot_results=True, plot_positions=True):
        Assesses the performance of the trading strategy on historical data.
    plot_results(title, plot_positions=True):
        Plots the performance of the trading strategy compared to a buy and hold strategy.
    _gen_repeating(s):
        A generator function that groups repeated elements in an iterable.
    plot_func(ax, group):
        A function used for plotting positions.
    """
    def __init__(
        self,
        symbol,
        amount,
        trading_costs,
        include_margin=False,
        leverage=1,
        margin_threshold=0.8,
        exchange='binance'
    ):
        """
        Initialize the BacktestMixin object.

        Parameters:
        -----------
        symbol : str
            The trading symbol to use for the backtest.
        amount : float
            The initial amount of capital to allocate for the backtest.
        trading_costs : float
            The transaction costs (e.g., spread, commissions) as a percentage.
        include_margin : bool, optional
            Flag indicating whether margin trading is included in the backtest. Default is False.
        leverage : float, optional
            The initial leverage to apply for margin trading. Default is 1.
        margin_threshold : float, optional
            The margin ratio threshold for margin call detection. Default is 0.8.
        exchange : str, optional
            The exchange to simulate the backtest on. Default is 'binance'.

        Raises:
        -------
        SymbolInvalid:
            If the specified trading symbol is not found in the leverage brackets.

        Notes:
        ------
        If `include_margin` is set to True, the leverage brackets for the specified symbol
        will be loaded.
        """
        self.exchange = exchange

        self.include_margin = False
        self.leverage_limits = [1, 125]

        self.set_leverage(leverage)
        self.set_margin_threshold(margin_threshold)

        self.amount = amount
        self.symbol = symbol
        self.tc = trading_costs / 100
        self.strategy = None
        self._original_data = None

        self.perf = 0
        self.outperf = 0
        self.results = None

        if include_margin:
            self.set_include_margin()

    def __getattr__(self, attr):
        """
        Overrides the __getattr__ method to get attributes from the trading strategy object.

        Parameters
        ----------
        attr : str
            The attribute to be retrieved.

        Returns
        -------
        object
            The attribute object.
        """
        try:
            method = getattr(self.strategy, attr)
            return method
        except AttributeError:
            return getattr(self, attr)

    def __repr__(self):
        extra_title = (f"<b>Initial Amount</b> = {self.amount} | "
                       f"<b>Trading Costs</b> = {self.tc * 100}% | "
                       f"<b>Leverage</b> = {self.leverage}")
        return extra_title + '<br>' + self.strategy.__repr__()

    def set_leverage(self, leverage):
        if isinstance(leverage, int) and self.leverage_limits[0] <= leverage <= self.leverage_limits[1]:
            self.leverage = leverage
        else:
            raise LeverageInvalid(leverage)

    def set_margin_threshold(self, margin_threshold):
        if isinstance(margin_threshold, (int, float)) and 0 < margin_threshold <= 1:
            self.margin_threshold = margin_threshold
        else:
            raise ValueError('Margin threshold must be between 0 and 1.')

    def set_include_margin(self):
        self.include_margin = True

        brackets = self._load_leverage_brackets()

        try:
            self._symbol_bracket = pd.DataFrame(brackets[self.symbol])
        except KeyError:
            raise SymbolInvalid(self.symbol)

    def load_data(self, data=None, csv_path=None):
        if data is None or csv_path:
            csv_path = csv_path if csv_path else config_vars.ohlc_data_file
            data = pd.read_csv(csv_path, index_col='date', parse_dates=True)
            data = data[~data.index.duplicated(keep='last')]  # remove duplicates

        self._original_data = data

        self.set_data(data.copy(), self.strategy)

    def run(self, print_results=True, plot_results=True, leverage=None):
        """Runs the trading strategy and prints and/or plots the results.

        Parameters:
        -----------
        params : dict or None
            The parameters to use for the trading strategy.
        print_results : bool
            If True, print the results of the backtest.
        plot_results : bool
            If True, plot the performance of the trading strategy compared to a buy and hold strategy.

        Returns:
        --------
        None
        """
        if leverage is not None:
            self.set_leverage(leverage)

        perf, outperf, results = self._test_strategy(print_results=print_results, plot_results=plot_results)

        self.perf = perf
        self.outperf = outperf
        self.results = results

    def optimize(self, params, **kwargs):
        """Optimizes the trading strategy using brute force.

        Parameters:
        -----------
        params : dict, list
            A dictionary or list (for strategy combintion) containing the parameters to optimize.
            The parameters must be given as the keywords of a dictionary, and the value is an array
            of the lower limit, upper limit and step, respectively.

            Example for single strategy:
                params = dict(window=(10, 20, 1))

            Example for multiple strategies
                params = [dict(window=(10, 20, 1)), dict(ma=(30, 50, 2)]
        **kwargs : dict
            Additional arguments to pass to the `brute` function.

        Returns:
        --------
        opt : numpy.ndarray
            The optimal parameter values.
        -self._update_and_run(opt, plot_results=True) : float
            The negative performance of the strategy using the optimal parameter values.
        """

        opt_params, strategy_params_mapping, optimization_steps = self._adapt_optimization_input(params)

        self.bar = progressbar.ProgressBar(max_value=optimization_steps, redirect_stdout=True)
        self.optimization_steps = 0

        opt = brute(
            self._update_and_run, opt_params,
            (False, False, strategy_params_mapping),
            finish=None,
            **kwargs
        )

        if not isinstance(opt, (list, tuple, type(np.array([])))):
            opt = np.array([opt])

        return (self._get_params_mapping(opt, strategy_params_mapping),
                -self._update_and_run(opt, True, True, strategy_params_mapping))

    def maximum_leverage(self, margin_threshold=None):
        """
        Find the maximum allowable leverage that keeps the margin ratio below a specified threshold.

        Parameters:
        -----------
        threshold: float, optional
            The desired maximum margin ratio threshold, must be between 0 and 1.

        Raises:
        -----------
        ValueError: If the provided threshold is not within the valid range (0, 1].

        Returns:
        -----------
        int:
            The maximum allowable leverage that satisfies the specified margin ratio threshold.
        """
        initial_leverage = self.leverage
        initial_margin_threshold = self.margin_threshold
        initial_include_margin = self.include_margin

        if margin_threshold is not None:
            self.set_margin_threshold(margin_threshold)

        self.set_include_margin()

        left_limit, right_limit = self.leverage_limits

        prev_leverage = 0

        print('\tCalculating maximum allowed leverage', end='')

        i = 0
        while True:
            leverage = math.floor((right_limit - left_limit) / 2) + left_limit

            if leverage == prev_leverage:
                break

            self._test_strategy(leverage=leverage, print_results=False, plot_results=False)
            if self.processed_data["margin_ratios"].max() >= self.margin_threshold:
                right_limit = leverage
            else:
                left_limit = leverage

            prev_leverage = leverage
            i += 1
            print('.', end='')

        logging.info('')

        self.set_leverage(initial_leverage)
        self.set_margin_threshold(initial_margin_threshold)
        self.include_margin = initial_include_margin

        return leverage

    def _test_strategy(self, params=None, leverage=None, print_results=True, plot_results=True, show_plot_no_tc=False):
        """Tests the trading strategy on historical data.

        Parameters:
        -----------
        params : dict or None
            The parameters to use for the trading strategy
        """
        if leverage is not None:
            self.set_leverage(leverage)

        self._fix_original_data()

        self.set_parameters(params, data=self._original_data.copy())

    @staticmethod
    def _get_optimization_input(optimization_params, strategy):
        opt_params = []
        optimizations_steps = 1
        for param in strategy.params:
            param_value = getattr(strategy, f"_{param}")
            is_int = isinstance(param_value, int)
            is_float = isinstance(param_value, float)

            step = 1 if is_int else None

            limits = optimization_params[param] \
                if param in optimization_params \
                else (param_value, param_value + 1) if is_int or is_float \
                else None

            if limits is not None:
                params = (*limits, step) if step is not None else limits
                opt_params.append(params)

                optimizations_steps *= (limits[1] - limits[0])

        return opt_params, optimizations_steps

    def _adapt_optimization_input(self, params):

        if not self.strategy:
            raise StrategyRequired

        if isinstance(self.strategy, StrategyCombiner):
            if not isinstance(params, (list, tuple, type(np.array([])))):
                raise OptimizationParametersInvalid('Optimization parameters must be provided as a list'
                                                    ' of dictionaries with the parameters for each individual strategy')

            if len(params) != len(self.strategy.strategies):
                raise OptimizationParametersInvalid(f'Wrong number of parameters. '
                                                    f'Number of strategies is {len(self.strategy.strategies)}')

            opt_params = []
            strategy_params_mapping = []
            optimization_steps = 1
            for i, strategy in enumerate(self.strategy.strategies):
                strategy_params, opt_steps = self._get_optimization_input(params[i], strategy)
                opt_params.extend(strategy_params)
                strategy_params_mapping.append(len(strategy_params))

                optimization_steps *= opt_steps

            return opt_params, strategy_params_mapping, optimization_steps

        else:
            if not isinstance(params, dict):
                raise OptimizationParametersInvalid('Optimization parameters must be provided as a '
                                                    'dictionary with the parameters the strategy')

            strategy_params, optimization_steps = self._get_optimization_input(params, self.strategy)

            return strategy_params, None, optimization_steps
        
    @staticmethod
    def _sanitize_margin_ratio(df):
        df["margin_ratios"] = np.where(df["margin_ratios"] > 1, 1, df["margin_ratios"])
        df["margin_ratios"] = np.where(df["margin_ratios"] < 0, 1, df["margin_ratios"])
        df["margin_ratios"] = np.where(df["side"] == 0, 0, df["margin_ratios"])

        df["margin_ratios"] = df["margin_ratios"].fillna(0)

        return df

    def _get_results(self, trades, processed_data):
        return get_results(
            processed_data,
            trades,
            self.leverage,
            self.amount,
            self.tc,
            config_vars.trading_days
        )

    @staticmethod
    def print_results(results, print_results):
        if not print_results:
            return

        log_results(results)

    def plot_results(self, processed_data, plot_results=True, show_plot_no_tc=False):
        """
        Plot the performance of the trading strategy compared to a buy and hold strategy.

        Parameters:
        -----------
        processed_data: pd.DataFrame
            Dataframe containing the results of the backtest to be plotted.
        plot_results: boolean, default True
            Whether to plot the results.
        show_plot_no_tc: boolean, default False
            Whether to show the plot of the equity curve with no trading costs
        """

        columns = [
            "accumulated_returns",
            "accumulated_strategy_returns",
            "accumulated_strategy_returns_tc",
        ]

        data = processed_data.copy()[columns] * self.amount

        if self.include_margin:
            data["margin_ratios"] = processed_data["margin_ratios"].copy() * 100

        trades_df = pd.DataFrame(self.trades)

        if plot_results:
            nr_strategies = len([col for col in processed_data.columns if "side" in col])
            offset = max(0, nr_strategies - 2)

            title = self.__repr__()

            plot_backtest_results(
                data,
                trades_df,
                self.margin_threshold,
                offset,
                plot_margin_ratio=self.include_margin,
                show_plot_no_tc=show_plot_no_tc,
                title=title
            )

    def _get_params_mapping(self, parameters, strategy_params_mapping):
        if not isinstance(self.strategy, StrategyCombiner):
            strategy_params = list(self.strategy.get_params().keys())
            new_params = {strategy_params[i]: parameter for i, parameter in enumerate(parameters)}
        else:
            new_params = []

            j = -1
            for i, mapping in enumerate(strategy_params_mapping):
                params = {}
                strategy_params = list(self.strategy.get_params(strategy_index=i).keys())
                for k, j in enumerate(range(j + 1, j + 1 + mapping)):
                    params.update({strategy_params[k]: parameters[j]})

                new_params.append(params)

        return new_params

    def _update_and_run(self, parameters, *args):
        """
        Update the hyperparameters of the strategy with the given `args`,
        and then run the strategy with the updated parameters.
        The strategy is run by calling the `_test_strategy` method with the
        updated parameters.

        Parameters
        ----------
        parameters : array-like
            A list of hyperparameters to be updated in the strategy.
            The order of the elements in the list should match the order
            of the strategy's hyperparameters, as returned by `self.params`.
        plot_results : bool, optional
            Whether to plot the results of the strategy after running it.

        Returns
        -------
        float
            The negative value of the strategy's score obtained with the
            updated hyperparameters. The negative value is returned to
            convert the maximization problem of the strategy's score into
            a minimization problem, as required by optimization algorithms.

        Raises
        ------
        IndexError
            If the number of elements in `parameters` does not match the number
            of hyperparameters in the strategy.

        Notes
        -----
        This method is intended to be used as the objective function to
        optimize the hyperparameters of the strategy using an optimization
        algorithm. It updates the hyperparameters of the strategy with the
        given `parameters`, then runs the strategy with the updated parameters,
        and returns the negative of the score obtained by the strategy.
        The negative is returned to convert the maximization problem of the
        strategy's score into a minimization problem, as required by many
        optimization algorithms.
        """
        print_results, plot_results, strategy_params_mapping = args

        test_params = self._get_params_mapping(parameters, strategy_params_mapping)

        result = self._test_strategy(test_params, print_results=print_results, plot_results=plot_results)

        self.optimization_steps += 1

        try:
            self.bar.update(self.optimization_steps)
        except ValueError:
            pass

        return -result[0]

    def _fix_original_data(self):
        if self._original_data is None:
            self._original_data = self.strategy.data.copy()

            position_columns = [col for col in self._original_data if "side" in col]

            self._original_data = self._original_data.drop(columns=position_columns)

    @staticmethod
    def _load_leverage_brackets():

        filepath = config_vars.leverage_brackets_file
        with open(filepath, 'r') as f:
            data = json.load(f)

        brackets = {symbol["symbol"]: symbol["brackets"] for symbol in data}

        return brackets



import logging
from collections import defaultdict
from datetime import datetime


class Trader:

    def __init__(self, amount, units=0):

        self.initial_balance = amount
        self.current_balance = amount
        self.units = units
        self.leverage = 1
        self.nr_trades = 0
        self.trades = []
        self.date_col = 'date'
        self.position = defaultdict(int)
        self.positions = defaultdict(list)
        self.strategy_returns = defaultdict(list)
        self.strategy_returns_tc = defaultdict(list)

    def _set_position(self, symbol, value, **kwargs):
        """
        Sets the side for the given symbol.

        Parameters
        ----------
        symbol : str
            The trading symbol.
        value : int
            The side value.
        """
        self.position[symbol] = value
        self.positions[symbol].append(value)

    def _reset_object(self, symbol):
        """
        Resets the object attributes to their initial values.
        """
        self.positions[symbol] = []
        self._set_position(symbol, 0)
        self.strategy_returns[symbol] = []
        self.strategy_returns_tc[symbol] = []

    def _get_position(self, symbol):
        raise NotImplementedError

    def buy_instrument(
        self,
        symbol,
        date=None,
        row=None,
        units=None,
        amount=None,
        header='',
        open_trade=False,
        **kwargs
    ):
        raise NotImplementedError

    def sell_instrument(
        self,
        symbol,
        date=None,
        row=None,
        units=None,
        amount=None,
        header='',
        open_trade=False,
        **kwargs
    ):
        raise NotImplementedError

    def close_pos(self, symbol, date=None, row=None, header='', **kwargs):
        raise NotImplementedError

    def _get_balances(self, symbol):

        initial_balance = self.initial_balance[symbol] if isinstance(self.initial_balance, dict) else self.initial_balance
        current_balance = self.current_balance[symbol] if isinstance(self.current_balance, dict) else self.current_balance

        return initial_balance, current_balance

    def _get_units(self, symbol):
        return self.units[symbol] if isinstance(self.units, dict) else self.units

    def go_long(self, symbol, position, date, row, units=None, amount=None, header='', **kwargs):

        existing_units = -self._get_units(symbol)

        if position == -1:
            self.buy_instrument(symbol, date, row, units=existing_units, header=header, reducing=True, **kwargs)

        if units:
            self.buy_instrument(symbol, date, row, units=units, header=header, open_trade=True, **kwargs)
        elif amount:
            if amount == "all":
                amount = self._get_balances(symbol)[1]
            self.buy_instrument(symbol, date, row, amount=amount, header=header, open_trade=True, **kwargs)  # go long

    # helper method
    def go_short(self, symbol, position, date, row, units=None, amount=None, header='', **kwargs):

        existing_units = self._get_units(symbol)

        if position == 1:
            self.sell_instrument(symbol, date, row, units=existing_units, header=header, reducing=True, **kwargs)

        if units:
            self.sell_instrument(symbol, date, row, units=units, header=header, open_trade=True, **kwargs)
        elif amount:
            if amount == "all":
                amount = self._get_balances(symbol)[1]
            self.sell_instrument(symbol, date, row, amount=amount, header=header, open_trade=True, **kwargs)  # go short

    def trade(self, symbol, signal, date=None, row=None, amount=None, units=None, header='', **kwargs):

        position = self._get_position(symbol)
        print_results = kwargs.get('print_results')

        if signal == 1:  # signal to go long
            if position in [0, -1]:
                # go long with full amount
                self.go_long(symbol, position, date, row, amount=amount, units=units, header=header, **kwargs)
        elif signal == -1:  # signal to go short
            if position in [0, 1]:
                # go short with full amount
                self.go_short(symbol, position, date, row, amount=amount, units=units, header=header, **kwargs)
        elif signal == 0:

            units = self.units[symbol] if isinstance(self.units, dict) else self.units

            if position == -1:
                self.buy_instrument(symbol, date, row, units=-units, header=header, reducing=True, **kwargs)
            elif position == 1:
                self.sell_instrument(symbol, date, row, units=units, header=header, reducing=True, **kwargs)

        if position == signal:
            if print_results:

                verbose_position = "LONG" if position == 1 else "SHORT" if position == -1 else "NEUTRAL"

                logging.info(header + f"Maintaining {verbose_position} position.")

                self.print_current_balance(date, header, symbol=symbol)

        self._set_position(symbol, signal, previous_position=position, **kwargs)

    def print_current_position_value(self, date, price, header='', **kwargs):
        cpv = self.get_current_position_value(price, **kwargs)
        logging.info(header + f"| {date} | Current Position Value = {round(cpv, 2)}")

    def print_current_nav(self, date, price, header='', **kwargs):
        nav = self.get_current_nav(price, **kwargs)
        logging.info(header + f"| {date} | Net Asset Value = {round(nav, 2)}")

    def print_current_balance(self, date, header='', **kwargs):
        symbol = kwargs.get("symbol", "")
        current_balance = self._get_balances(symbol)[1]

        logging.info(header + f"| {date if date else datetime.now()} | Current Balance: {round(current_balance, 2)}")

    def get_current_position_value(self, price, **kwargs):
        units = self._get_units(kwargs.get("symbol", ""))

        return units * price

    def get_current_nav(self, price, **kwargs):
        symbol = kwargs.get("symbol", "")
        current_balance = self._get_balances(symbol)[1]
        current_position_value = self.get_current_position_value(price, **kwargs)

        return current_balance + current_position_value



from collections import OrderedDict

import numpy as np

from stratestic.strategies._mixin import StrategyMixin


class BollingerBands(StrategyMixin):
    """ Bollinger Bands Strategy:

    This strategy follows the principle of mean reversion, ie. that
    the price will revert back to the mean if it deviates by a certain amount,
    essentially going long or short when the price crosses a
    certain low and high threshold respectively.

    Parameters:
    -----------
    ma : int
        Moving average window.
    sd : int
        Standard deviation window.
    data : pd.DataFrame, default None
        Data to use in the strategy.
    **kwargs
        Additional arguments to pass to the `StrategyMixin` superclass.

    Attributes:
    -----------
    params : OrderedDict
        Ordered dictionary containing the strategy's parameters:
        - `ma`: moving average window.
        - `sd`: standard deviation window.

    Methods:
    --------
    __repr__(self)
        Return a string representation of the class instance.
    update_data(self)
        Retrieves and prepares the data.
    calculate_positions(self, data)
        Calculate the side for each row in the data.
    _get_position(self, symbol)
        Return the side for a given symbol (not implemented).
    get_signal(self, row=None)
        Return the side signal for a given row.

    """

    def __init__(self, ma: int, sd: int, data=None, **kwargs):

        self._ma = ma
        self._sd = sd

        StrategyMixin.__init__(self, data, **kwargs)

        self.params = OrderedDict(
            ma=lambda x: int(x),
            sd=lambda x: float(x),
        )

    def __repr__(self):
        return "{}(symbol = {}, ma = {}, sd = {})".format(self.__class__.__name__, self.symbol, self._ma, self._sd)

    def update_data(self, data):
        """
        Updates the input data with additional columns required for the strategy.

        Parameters
        ----------
        data : pd.DataFrame
            OHLCV data to be updated.

        Returns
        -------
        pd.DataFrame
            Updated OHLCV data containing additional columns.
        """
        data = super().update_data(data)

        data["sma"] = data[self.close_col].rolling(self._ma).mean()
        data["upper"] = data["sma"] + data[self.close_col].rolling(self._ma).std() * self._sd
        data["lower"] = data["sma"] - data[self.close_col].rolling(self._ma).std() * self._sd

        return self.calculate_positions(data)

    def calculate_positions(self, data):
        data["distance"] = data[self.close_col] - data["sma"]
        data["side"] = np.where(data[self.close_col] > data["upper"], -1, np.nan)
        data["side"] = np.where(data[self.close_col] < data["lower"], 1, data["side"])
        data["side"] = np.where(data["distance"] * data["distance"].shift(1) < 0, 0, data["side"])
        data["side"] = data["side"].ffill().fillna(0)

        return data

    def _get_position(self, symbol):
        return None

    def get_signal(self, row=None):

        if row is None:
            row = self.data.iloc[-1]

        return int(row["side"])



from collections import OrderedDict
from typing import Literal

import numpy as np
from ta.trend import ema_indicator, sma_indicator

from stratestic.strategies._mixin import StrategyMixin


class MovingAverageCrossover(StrategyMixin):
    """
    A strategy that uses two simple or exponential moving averages as a trigger for going long or short, with no neutral
    side. When the shorter moving average goes above the longer moving average it goes long, and goes short when the
    reverse holds true.

    Parameters:
    -----------
    sma_s: int
        The window size for the short moving average.
    sma_l: int
        The window size for the long moving average.
    moving_av: Literal["sma", "ema"], default='sma'
        The type of moving average to use. Either 'sma' for simple moving average, or 'ema' for exponential moving average.

    Attributes:
    -----------
    params : OrderedDict
        A dictionary of the parameters used to initialize the strategy.

    Methods:
    --------
    update_data():
        Retrieves and prepares the data for the strategy.

    calculate_positions(data):
        Calculates the side values for the given data.

    get_signal(row=None):
        Gets the trading signal for a given row of data. If row is not specified, uses the last row of data.

    """

    def __init__(
        self,
        sma_s: int,
        sma_l: int,
        moving_av: Literal["sma", "ema"] = 'sma',
        data=None,
        **kwargs
    ):
        """
        Constructs a new instance of the MovingAverageCrossover class.

        Parameters:
        -----------
        sma_s: int
            The window size for the short moving average.
        sma_l: int
            The window size for the long moving average.
        moving_av: Literal["sma", "ema"], default='sma'
            The type of moving average to use. Either 'sma' for simple moving average, or 'ema' for exponential moving average.
        data: pandas.DataFrame, default=None
            The data to use for the strategy. If not specified, the data will be retrieved from the data source.
        kwargs:
            Additional keyword arguments to be passed to the parent class constructor.

        """
        self._sma_s = sma_s
        self._sma_l = sma_l
        self._moving_av = moving_av

        self.params = OrderedDict(sma_s=lambda x: int(x), sma_l=lambda x: int(x), moving_av=lambda x: x)

        StrategyMixin.__init__(self, data, **kwargs)

    def __repr__(self):
        """
        Returns a string representation of the MovingAverageCrossover instance.

        Returns:
        --------
        str:
            A string representation of the MovingAverageCrossover instance.
        """
        return "{}(symbol = {}, SMA_S = {}, SMA_L = {})".format(self.__class__.__name__, self.symbol, self._sma_s, self._sma_l)

    def update_data(self, data):
        """
        Updates the input data with additional columns required for the strategy.

        Parameters
        ----------
        data : pd.DataFrame
            OHLCV data to be updated.

        Returns
        -------
        pd.DataFrame
            Updated OHLCV data containing additional columns.
        """
        data = super().update_data(data)

        if self._moving_av == 'sma':
            data["SMA_S"] = sma_indicator(close=data[self.close_col], window=self._sma_s)
            data["SMA_L"] = sma_indicator(close=data[self.close_col], window=self._sma_l)

        elif self._moving_av == 'ema':
            data["SMA_S"] = ema_indicator(close=data[self.close_col], window=self._sma_s)
            data["SMA_L"] = ema_indicator(close=data[self.close_col], window=self._sma_l)
        else:
            raise ('Method not supported')

        return data

    def calculate_positions(self, data):
        """
        Calculates the side values for the given data.

        Parameters:
        -----------
        data: pandas.DataFrame
            The data to use for the calculation.

        Returns:
        --------
        pandas.DataFrame:
            The input DataFrame with an additional 'side' column containing the calculated side values.
        """
        data["side"] = np.where(data["SMA_S"] > data["SMA_L"], 1, -1)

        return data

    def get_signal(self, row=None):
        """
        Gets the trading signal for a given row of data.

        Parameters:
        -----------
        row: pandas.Series, default=None
            The row of data for which to get the trading signal. If not specified, the last row of data will be used.

        Returns:
        --------
        int:
            1 if the short moving average is above the long moving average, -1 if the short moving average is below the
            long moving average.
        """
        if row is None:
            row = self.data.iloc[-1]

        if row["SMA_S"] > row["SMA_L"]:
            return 1
        elif row["SMA_S"] < row["SMA_L"]:
            return -1
        else:
            return 0
